using Godot;

namespace Quadratic.Carto.MathExt;

public struct Ray
{
    public Vector3 origin;
    public Vector3 direction;

    /// <summary>
    /// Creates a new ray, using origin and <b>normalized</b> direction.
    /// </summary>
    /// <param name="origin">The origin vector.</param>
    /// <param name="direction">The direction vector, normalized.</param>
    public Ray(Vector3 origin, Vector3 direction)
    {
        this.origin = origin;
        this.direction = direction;
    }

    /// <summary>
    /// Creates a new ray, using start and end points.
    /// </summary>
    /// <param name="start">The start point vector.</param>
    /// <param name="end">The end point vector.</param>
    public static Ray FromStartAndEnd(Vector3 start, Vector3 end)
    {
        return new Ray(start, (end - start).Normalized());
    }

    /// <summary>
    /// Creates a new ray, using origin and non-normalized direction.
    /// </summary>
    /// <param name="origin">The origin vector.</param>
    /// <param name="direction">The direction vector, non-normalized.</param>
    public static Ray FromOriginAndDirection(Vector3 origin, Vector3 direction)
    {
        return new Ray(origin, direction.Normalized());
    }

    /// <summary>
    /// Creates a new ray, using origin and normalized direction.
    /// </summary>
    /// <param name="origin">The origin vector.</param>
    /// <param name="direction">The direction vector, normalized.</param>
    public static Ray FromOriginAndNormalizedDirection(Vector3 origin, Vector3 direction)
    {
        return new Ray(origin, direction);
    }

    /// <summary>
    /// Creates a new ray, using origin and rotation quaternion.
    /// </summary>
    /// <param name="origin">The origin vector.</param>
    /// <param name="rotation">The rotation quaternion.</param>
    public static Ray FromOriginAndRotation(Vector3 origin, Quaternion rotation)
    {
        return new Ray(origin, rotation * Vector3.Forward);
    }

    /// <summary>
    /// Get the point at the given distance along the ray.
    /// </summary>
    /// <param name="distance">The distance along the ray.</param>
    public readonly Vector3 GetPoint(float distance)
    {
        return this.origin + this.direction * distance;
    }

    /// <summary>
    /// Returns the point along the ray that is closest to the given point.
    /// </summary>
    /// <param name="point"></param>
    /// <returns></returns>
    public readonly float MinDistancePoint(Vector3 point)
    {
        /*
             The minimum distance between a ray and a point:

                                  d: direction
             ------+--------------<----* O: origin
                   |                    \
                   * P                   * P'

             If the line between origin and target point (OP) is on the same
             side as d, the distance is the magnitude of the vertical line
             between P and Od. Else (OP'), the distance is simply the distance 
             between O and P.
         */

        var op = point - this.origin; // vector OP
        var opd = op.Dot(this.direction); // OP dot d is the projection of OP onto d
        if (opd < 0) // OP' case
        {
            return op.Length();
        }
        else // OP case
        {
            var od = this.direction * opd; // vector Od
            return (op - od).Length();
        }
    }

    public readonly Vector3 ClosestPoint(Vector3 point)
    {
        var pointDist = MinDistancePoint(point);
        return GetPoint(pointDist);
    }

    public readonly Ray Transform(Transform3D transform)
    {
        return new Ray(
            transform.Origin + transform.Basis * this.origin,
            (transform.Basis * this.direction).Normalized()
        );
    }

    public readonly Ray InverseTransform(Transform3D transform)
    {
        var inv = transform.Inverse();
        return Transform(inv);
    }

    public override readonly string ToString()
    {
        return $"Ray({this.origin}, {this.direction})";
    }
}
