using Godot;

namespace Quadratic.Carto.MathExt;

public static partial class VectorsExt
{
    public static Quaternion FromToRotation(Vector3 from, Vector3 to)
    {
        var fromDotTo = from.Dot(to);
        if (Mathf.IsEqualApprox(fromDotTo, 1))
        {
            return Quaternion.Identity;
        }
        else if (Mathf.IsEqualApprox(fromDotTo, -1))
        {
            return new Quaternion(Vector3.Right, Mathf.Pi);
        }
        var axis = from.Cross(to).Normalized();
        var angle = Mathf.Acos(fromDotTo);
        return new Quaternion(axis, angle);
    }

    /// <summary>
    /// Calculates the closest distance between two lines, defined by origin and direction. This
    /// method assumes normalized direction vectors.
    /// </summary>
    /// <param name="aOrigin"></param>
    /// <param name="aDirection"></param>
    /// <param name="bOrigin"></param>
    /// <param name="bDirection"></param>
    public static float ClosestDistanceBetweenLines(
        Vector3 aOrigin,
        Vector3 aDirection,
        Vector3 bOrigin,
        Vector3 bDirection)
    {
        // Rule out the case where the lines are parallel.
        float aDotB = aDirection.Dot(bDirection);
        if (Mathf.IsEqualApprox(aDotB, 1))
        {
            // Lines are parallel
            var aToB = bOrigin - aOrigin;
            var aPerp = aToB - aDirection * aToB.Dot(aDirection);
            return aPerp.Length();
        }
        else
        {
            // The vector that's perpendicular to both lines.
            // Since both lines are normalized, this vector is also normalized.
            var normalAB = aDirection.Cross(bDirection);
            var aToB = bOrigin - aOrigin;
            return aToB.Dot(normalAB);
        }
    }

    /// <summary>
    /// Calculates the closest distance and points between two lines, defined by origin and
    /// direction. This method assumes normalized direction vectors.
    /// </summary>
    /// <param name="aOrigin"></param>
    /// <param name="aDirection"></param>
    /// <param name="bOrigin"></param>
    /// <param name="bDirection"></param>
    /// <param name="aClosest"></param>
    /// <param name="bClosest"></param>
    public static float ClosestDistanceBetweenLines(
        Vector3 aOrigin,
        Vector3 aDirection,
        Vector3 bOrigin,
        Vector3 bDirection,
        out Vector3 aClosest,
        out Vector3 bClosest)
    {
        // Rule out the case where the lines are parallel.
        float aDotB = aDirection.Dot(bDirection);
        if (Mathf.IsEqualApprox(aDotB, 1))
        {
            // Lines are parallel
            var aToB = bOrigin - aOrigin;
            var aPerp = aToB - aDirection * aToB.Dot(aDirection);
            aClosest = aOrigin;
            bClosest = bOrigin + aPerp;
            return aPerp.Length();
        }
        else
        {
            // https://en.wikipedia.org/wiki/Skew_lines#Nearest_Points
            // The vector that's perpendicular to both lines.
            // Since both lines are normalized, this vector is also normalized.
            var n = aDirection.Cross(bDirection);
            var n1 = aDirection.Cross(n);
            var n2 = bDirection.Cross(n);
            var aToB = bOrigin - aOrigin;
            aClosest = aOrigin + aToB.Dot(n2) / aDirection.Dot(n2) * aDirection;
            bClosest = bOrigin + aToB.Dot(n1) / bDirection.Dot(n1) * bDirection;
            return aToB.Dot(n);
        }
    }
}
